part of model_notifier;

class LocalTransaction {
  const LocalTransaction._();

  /// Save the number of elements in [collectionPath] to its parent document with [counterSuffix].
  ///
  /// You can add the corresponding element by specifying [linkedCollectionPath].
  ///
  /// You can generate a key to store the number of elements in a document by specifying [counterBuilder] or [linkedCounterBuilder].
  static LocalIncrementCounterTransactionBuilder incrementCounter({
    required String collectionPath,
    String counterSuffix = "Count",
    String Function(String path)? counterBuilder,
    String? linkedCollectionPath,
    String Function(String linkPath)? linkedCounterBuilder,
    List<CounterUpdaterInterval> counterIntervals = const [],
  }) {
    return LocalIncrementCounterTransactionBuilder._(
      collectionPath: collectionPath,
      counterSuffix: counterSuffix,
      counterBuilder: counterBuilder,
      linkedCollectionPath: linkedCollectionPath,
      linkedCounterBuilder: linkedCounterBuilder,
      counterIntervals: counterIntervals,
    );
  }

  /// Create a code of length [length] randomly for id.
  ///
  /// Characters that are difficult to understand are omitted.
  ///
  /// If the data of [key] in [path] contains the generated random value,
  /// the random value is generated again.
  static Future<String> generateCode({
    required String path,
    required String key,
    int length = 6,
    String charSet = "23456789abcdefghjkmnpqrstuvwxy",
  }) async {
    List<String> generated = [];
    do {
      // await Future.delayed(Duration(milliseconds: Random().nextInt(100)));
      generated = List.generate(
          10, (index) => katana.generateCode(length, charSet: charSet));
      final snapshot = await LocalDatabase._db.loadDocument(
        _LocalStoreDocumentQuery(path: path),
      );
      if (snapshot != null) {
        for (final map in snapshot.values) {
          if (map.isEmpty || !map.containsKey(key)) {
            continue;
          }
          generated.remove(map.get(key, ""));
        }
      }
    } while (generated.isEmpty);
    return generated.first;
  }
}

/// Generates a counter builder.
class LocalIncrementCounterTransactionBuilder
    extends IncrementCounterTransactionBuilder {
  LocalIncrementCounterTransactionBuilder._({
    required this.collectionPath,
    this.counterBuilder,
    this.counterSuffix = "Count",
    this.linkedCollectionPath,
    this.linkedCounterBuilder,
    this.counterIntervals = const [],
  })  : assert(
            !(collectionPath.splitLength() <= 0 ||
                collectionPath.splitLength() % 2 != 1),
            "The collection path hierarchy must be an odd number."),
        assert(
            linkedCollectionPath.isEmpty ||
                !(linkedCollectionPath!.splitLength() <= 0 ||
                    linkedCollectionPath.splitLength() % 2 != 1),
            "The link collection path hierarchy must be an odd number.");

  /// Suffix to generate the counter key.
  final String counterSuffix;

  /// A collection that counts the number of elements.
  final String collectionPath;

  /// Builder to generate the counter key.
  final String Function(String path)? counterBuilder;

  /// A linked collection that counts the number of elements.
  final String? linkedCollectionPath;

  /// Builder to generate the counter key for linked collection.
  final String Function(String linkPath)? linkedCounterBuilder;

  /// List of counters with valid expiration dates.
  final List<CounterUpdaterInterval> counterIntervals;

  String? _buildCounterPath(String? path) {
    if (path.isEmpty) {
      return null;
    }
    return "$path$counterSuffix";
  }

  DynamicMap _buildCounterUpdate(DynamicMap map, String key, num value) {
    final now = DateTime.now();
    map[key] = map.get(key, 0) + value;
    for (final interval in counterIntervals) {
      switch (interval) {
        case CounterUpdaterInterval.daily:
          final k = dailyKey(key, now);
          map[k] = map.get(k, 0) + value;
          for (var i = 0; i < 30; i++) {
            map[dailyKey(
                key, DateTime(now.year, now.month, now.day - 60 + i))] = null;
          }
          break;
        case CounterUpdaterInterval.monthly:
          final k = monthlyKey(key, now);
          map[k] = map.get(k, 0) + value;
          for (var i = 0; i < 12; i++) {
            map[monthlyKey(key, DateTime(now.year, now.month - 24 + i))] = null;
          }
          break;
        case CounterUpdaterInterval.yearly:
          final k = yearlyKey(key, now);
          map[k] = map.get(k, 0) + value;
          for (var i = 0; i < 5; i++) {
            map[yearlyKey(key, DateTime(now.year, now.month - 10 + i))] = null;
          }
          break;
        case CounterUpdaterInterval.weekly:
          final k = weeklyKey(key, now);
          map[k] = map.get(k, 0) + value;
          for (var i = 0; i < 4; i++) {
            map[weeklyKey(key,
                DateTime(now.year, now.month, now.day - ((8 - i) * 7)))] = null;
          }
          break;
      }
    }
    return map;
  }

  /// Add an element of [id] to [collectionPath].
  ///
  /// If [linkedCollectionPath] is specified, the element of [linkId] is added.
  @override
  Future<void> append(String id, {String? linkId}) async {
    assert(
        linkId.isEmpty ||
            (linkedCollectionPath.isNotEmpty && linkId.isNotEmpty),
        "When [linkId] is specified, [linkPath] must be specified.");
    // await Future.delayed(Duration(milliseconds: Random().nextInt(100)));
    final docPath = counterBuilder?.call(collectionPath) ??
        _buildCounterPath(collectionPath) ??
        "";
    final linkDocPath = linkedCollectionPath.isEmpty
        ? null
        : (linkedCounterBuilder?.call(linkedCollectionPath!) ??
            _buildCounterPath(linkedCollectionPath!) ??
            "");
    final updated = <String, DynamicMap>{};
    final inserted = <String, DynamicMap>{};
    try {
      final doc = await LocalDatabase._db.loadDocument(
        _LocalStoreDocumentQuery(
          path: "$collectionPath/$id",
        ),
      );
      final linkDoc = linkId.isEmpty || linkedCollectionPath.isEmpty
          ? null
          : await LocalDatabase._db.loadDocument(
              _LocalStoreDocumentQuery(
                path: "$linkedCollectionPath/$linkId",
              ),
            );
      if (doc.isEmpty) {
        inserted["$collectionPath/$id"] = {
          Const.uid: id,
          Const.time: DateTime.now().millisecondsSinceEpoch,
        };
        if (docPath.isNotEmpty) {
          final key = docPath.split("/").last;
          final path = docPath.replaceFirst(RegExp("/$key\$"), "");
          final countDoc = await LocalDatabase._db.loadDocument(
            _LocalStoreDocumentQuery(path: path),
          );
          updated[path] = _buildCounterUpdate(countDoc ?? {}, key, 1);
        }
      }
      if (linkId.isNotEmpty &&
          linkedCollectionPath.isNotEmpty &&
          linkDoc.isEmpty) {
        inserted["$linkedCollectionPath/$linkId"] = {
          Const.uid: linkId,
          Const.time: DateTime.now().millisecondsSinceEpoch,
        };
        if (linkDocPath.isNotEmpty) {
          final key = linkDocPath!.split("/").last;
          final path = linkDocPath.replaceFirst(RegExp("/$key\$"), "");
          final countDoc = await LocalDatabase._db.loadDocument(
            _LocalStoreDocumentQuery(path: path),
          );
          updated[path] = _buildCounterUpdate(countDoc ?? {}, key, 1);
        }
      }
      for (final tmp in updated.entries) {
        LocalDatabase._db.saveDocument(
          _LocalStoreDocumentQuery(
            path: tmp.key,
          ),
          tmp.value,
        );
      }
      for (final tmp in inserted.entries) {
        LocalDatabase._db.saveDocument(
          _LocalStoreDocumentQuery(
            path: tmp.key,
          ),
          tmp.value,
        );
      }
    } catch (e) {
      print(e.toString());
      rethrow;
    }
  }

  /// Remove an element of [id] to [collectionPath].
  ///
  /// If [linkedCollectionPath] is specified, the element of [linkId] is removed.
  @override
  Future<void> remove(String id, {String? linkId}) async {
    assert(
        linkId.isEmpty ||
            (linkedCollectionPath.isNotEmpty && linkId.isNotEmpty),
        "When [linkId] is specified, [linkPath] must be specified.");
    final docPath = counterBuilder?.call(collectionPath) ??
        _buildCounterPath(collectionPath) ??
        "";
    final linkDocPath = linkedCollectionPath.isEmpty
        ? null
        : (linkedCounterBuilder?.call(linkedCollectionPath!) ??
            _buildCounterPath(linkedCollectionPath!) ??
            "");
    final updated = <String, DynamicMap>{};
    final deleted = <String>[];
    try {
      final doc = await LocalDatabase._db.loadDocument(
        _LocalStoreDocumentQuery(
          path: "$collectionPath/$id",
        ),
      );
      final linkDoc = linkId.isEmpty || linkedCollectionPath.isEmpty
          ? null
          : await LocalDatabase._db.loadDocument(
              _LocalStoreDocumentQuery(
                path: "$linkedCollectionPath/$linkId",
              ),
            );
      if (doc.isNotEmpty) {
        deleted.add("$collectionPath/$id");
        if (docPath.isNotEmpty) {
          final key = docPath.split("/").last;
          final path = docPath.replaceFirst(RegExp("/$key\$"), "");
          final countDoc = await LocalDatabase._db.loadDocument(
            _LocalStoreDocumentQuery(path: path),
          );
          updated[path] = _buildCounterUpdate(countDoc ?? {}, key, -1);
        }
      }
      if (linkId.isNotEmpty &&
          linkedCollectionPath.isNotEmpty &&
          linkDoc.isNotEmpty) {
        deleted.add("$linkedCollectionPath/$linkId");
        if (linkDocPath.isNotEmpty) {
          final key = linkDocPath!.split("/").last;
          final path = linkDocPath.replaceFirst(RegExp("/$key\$"), "");
          final countDoc = await LocalDatabase._db.loadDocument(
            _LocalStoreDocumentQuery(path: path),
          );
          updated[path] = _buildCounterUpdate(countDoc ?? {}, key, -1);
        }
      }
      for (final tmp in updated.entries) {
        LocalDatabase._db.saveDocument(
          _LocalStoreDocumentQuery(
            path: tmp.key,
          ),
          tmp.value,
        );
      }
      for (final tmp in deleted) {
        LocalDatabase._db.deleteDocument(
          _LocalStoreDocumentQuery(
            path: tmp,
          ),
        );
      }
    } catch (e) {
      print(e.toString());
      rethrow;
    }
  }
}
